@@ -230,7 +230,7 @@ void run_program(void)
     if (commandline_print)
     {
       std::cout << "\nHex dump of inode " << commandline_inode << ":\n";
-      dump_hex_to(std::cout, (unsigned char const*)&inode, inode_size_);
+      dump_hex_to(std::cout, (unsigned char const*)&(*inode), inode_size_);
       std::cout << '\n';
     }
     unsigned int bit = commandline_inode - 1 - commandline_group * inodes_per_group_;
@@ -387,9 +387,28 @@ void run_program(void)
 	else
 	{
 	  std::cout << "Block " << commandline_block << " is Unallocated.\n";
-	  ASSERT(!is_inode(commandline_block));	// All inode blocks are allocated.
+	  // If this assertion fails, then it is possible that this DATA block looks like an inode,
+	  // most likely because the data itself is an ext3 filesystem. For example an ext3 image.
+	  // If that is possible, then just comment this assertion out.
+	  //ASSERT(!is_inode(commandline_block));	// All inode blocks are allocated.
 	  ASSERT(!journal);			// All journal blocks are allocated.
 	}
+	if (is_indirect_block(block))
+	{
+	  std::cout << "Block " << commandline_block << " appears to be an (double/tripple) indirect block.\n";
+	  if (commandline_print)
+	  {
+	    std::cout << "It contains the following block numbers:\n";
+	    __le32* block_numbers = reinterpret_cast<__le32*>(block);
+	    for (int i = 0; i < block_size_ >> 2; ++i)
+	    {
+	      std::cout << ' ' << std::setw(9) << std::setfill(' ') << block_numbers[i];
+	      if ((i + 1) % 10 == 0)
+	        std::cout << '\n';
+	    }
+	    std::cout << '\n';
+	  }
+	}
       }
       else
       {
@@ -468,7 +487,8 @@ void run_program(void)
     dump_names();
   // Handle --restore-file
   if (!commandline_restore_file.empty())
-    restore_file(commandline_restore_file);
+    for (std::vector<std::string>::iterator iter = commandline_restore_file.begin(); iter != commandline_restore_file.end(); ++iter)
+      restore_file(*iter);
   // Handle --restore-inode
   if (!commandline_restore_inode.empty())
   {
